# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from mistral_lib.actions import base


class CompositeActionProvider(base.ActionProvider):
    def __init__(self, name, delegates):
        super().__init__(name)

        self._delegates = delegates

    def find(self, action_name, namespace=None):
        for d in self._delegates:
            action_desc = d.find(action_name, namespace)

            if action_desc is not None:
                return action_desc

        return None

    def find_all(self, namespace=None, limit=None, sort_fields=None,
                 sort_dirs=None, **filters):
        # TODO(rakhmerov): Implement the algorithm that takes ordering/sorting
        # parameters into account correctly. For now, they are just passed to
        # delegates.
        res = []

        for d in self._delegates:
            action_descriptors = d.find_all(
                namespace=namespace,
                limit=limit,
                sort_fields=sort_fields,
                sort_dirs=sort_dirs,
                **filters
            )

            if action_descriptors is not None:
                res.extend(action_descriptors)

        return res

    def add_action_provider(self, action_provider):
        self._delegates.append(action_provider)



# Copyright 2016 - Nokia Networks.
# Copyright 2017 - Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import abc

from oslo_utils import importutils

from mistral_lib import serialization
from mistral_lib.utils import inspect_utils as i_utils


class Action(serialization.MistralSerializable):
    """Action.

    Action is a means in Mistral to perform some useful work associated with
    a workflow during its execution. Every workflow task is configured with
    an action and when the task runs it eventually delegates to the action.
    When it happens task parameters get evaluated (calculating expressions,
    if any) and are treated as action parameters. So in a regular general
    purpose languages terminology action is a method declaration and task is
    a method call.

    Base action class initializer doesn't have arguments. However, concrete
    action classes may have any number of parameters defining action behavior.
    These parameters must correspond to parameters declared in action
    specification (e.g. using DSL or others).
    """

    def __init__(self):
        # NOTE(d0ugal): We need to define an empty __init__ otherwise
        # inspect.getargspec will fail in Python 2 for actions that subclass
        # but don't define their own __init__.
        pass

    @abc.abstractmethod
    def run(self, context):
        """Run action logic.

        :param context: contains contextual information of the action.
        The context includes an execution context (like execution identifier
        and workflow name) and a security context with the authorization
        details.
        :return: Result of the action. Note that for asynchronous actions
        it should always be None, however, if even it's not None it will be
        ignored by a caller.

        Result can be of two types:
        1) Any serializable value meaningful from a user perspective (such
        as string, number or dict).
        2) Instance of {mistral_lib.types.Result} which has field "data"
        for success result and field "error" for keeping so called "error
        result" like HTTP error code and similar. Using the second type
        allows to communicate a result even in case of error and hence to have
        conditions in "on-error" clause of direct workflows. Depending on
        particular action semantics one or another option may be preferable.
        In case if action failed and there's no need to communicate any error
        result this method should throw a ActionException.
        """
        pass

    def is_sync(self):
        """Returns True if the action is synchronous, otherwise False.

        :return: True if the action is synchronous and method run() returns
            final action result. Otherwise returns False which means that
            a result of method run() should be ignored and a real action
            result is supposed to be delivered in an asynchronous manner
            using public API. By default, if a concrete implementation
            doesn't override this method then the action is synchronous.
        """
        return True

    @classmethod
    def get_serialization_key(cls):
        # By default we use the same serializer key for every action
        # assuming that most action class can use the same serializer.
        return "%s.%s" % (Action.__module__, Action.__name__)


class ActionSerializer(serialization.DictBasedSerializer):
    def serialize_to_dict(self, entity):
        cls = type(entity)

        return {
            'cls': '%s.%s' % (cls.__module__, cls.__name__),
            'cls_attrs': i_utils.get_public_fields(cls),
            'data': vars(entity),
        }

    def deserialize_from_dict(self, entity_dict):
        cls_str = entity_dict['cls']

        # Rebuild action class and restore attributes.
        cls = importutils.import_class(cls_str)

        cls_attrs = entity_dict['cls_attrs']

        if cls_attrs:
            # If we have serialized class attributes it means that we need
            # to create a dynamic class.
            cls = type(cls.__name__, (cls,), cls_attrs)

        # NOTE(rakhmerov): We use this hacky was of instantiating
        # the action here because we can't use normal __init__(),
        # we don't know the parameters. And even if we find out
        # what they are the real internal state of the object is
        # what was stored as vars() method that just returns all
        # fields. So we have to bypass __init__() and set attributes
        # one by one. Of course, this is a serious limitation since
        # action creators will need to keep in mind to avoid having
        # some complex initialisation logic in __init__() that
        # does something not reflecting in an instance state.
        # However, this all applies to the case when the action
        # has to be sent to a remote executor.
        action = cls.__new__(cls)

        for k, v in entity_dict['data'].items():
            setattr(action, k, v)

        return action


# NOTE: Every action implementation can register its own serializer
# if needed, but this serializer should work for vast of majority of
# actions.
serialization.register_serializer(Action, ActionSerializer())


class ActionDescriptor(abc.ABC):
    """Provides required information about a certain type of actions.

    Action descriptor is not an action itself. It rather carries all
    important meta information about a particular action before the
    action is instantiated. In some sense it is similar to a class but
    the difference is that one type of action descriptor may support
    many different actions. This abstraction is responsible for
    validation of input parameters and instantiation of an action.

    """

    @property
    @abc.abstractmethod
    def name(self):
        """The name of the action."""
        pass

    @property
    @abc.abstractmethod
    def description(self):
        """The description of the action."""
        pass

    @property
    @abc.abstractmethod
    def params_spec(self):
        """Comma-separated string with input parameter names.

        Each parameter name can be either just a name or a string
        "param=val" where "param" is the name of the parameter
        and "val" its default value. The values are only indications
        for the user and not used in the action instantiation process.
        The string is split along the commas and then the parts along the
        equal signs. Escaping is not possible.
        """
        pass

    @property
    @abc.abstractmethod
    def namespace(self):
        """The namespace of the action.

        NOTE: Not all actions have to support namespaces.
        """
        pass

    @property
    @abc.abstractmethod
    def project_id(self):
        """The ID of the project (tenant) this action belongs to.

        If it's not specified then the action can be used within
        all projects (tenants).

        NOTE: Not all actions have to support projects(tenants).
        """
        pass

    @property
    @abc.abstractmethod
    def scope(self):
        """The scope of the action within a project (tenant).

        It makes sense only if the "project_id" property is not None.
        It should be assigned with the "public" value if the action
        is available in all projects and "private" if it's accessible
        only by users of the specified project.

        NOTE: Not all actions have to support projects(tenants).
        """
        pass

    @property
    @abc.abstractmethod
    def action_class_name(self):
        """String representation of the Python class of the action.

        Can be None in case if the action is dynamically generated with
        some kind of wrapper.
        """
        pass

    @property
    @abc.abstractmethod
    def action_class_attributes(self):
        """The attributes of the action Python class, if relevant.

        If the action has a static Python class associated with it
        and this method returns not an empty dictionary then the
        action can be instantiated from a new dynamic class
        based on the property "action_class_string" and this property
        that essentially carries public class field values.
        """
        pass

    @abc.abstractmethod
    def instantiate(self, input_dict, wf_ctx):
        """Instantiate the required action with the given parameters.

        :param input_dict: Action parameters as a dictionary where keys
            are parameter names and values are parameter values.
        :param wf_ctx: Workflow context relevant for the point when
            action is about to start.
        :return: An instance of mistral_lib.actions.Action.
        """
        pass

    @abc.abstractmethod
    def check_parameters(self, params):
        """Validate action parameters.

        The method does a preliminary check of the given actual action
        parameters and raises an exception if they don't match the
        action signature. However, a successful invocation of this
        method does not guarantee a further successful run of the
        instantiated action.

        :param params: Action parameters as a dictionary where keys
            are parameter names and values are parameter values.
        :return: None or raises an exception if the given parameters
            are not valid.
        """
        pass

    @abc.abstractmethod
    def post_process_result(self, result):
        """Converts the given action result.

        A certain action implementation may need to do an additional
        conversion of the action result by its descriptor. This approach
        allows to implement wrapper actions running asynchronously
        because in such cases, the initial action result depends on a 3rd
        party that's responsible for delivering it to Mistral. But when
        it comes to Mistral we still have a chance to apply needed
        transformations defined by this method.

        :param result: Action result.
            An instance of mistral_lib.types.Result.
        :return: Converted action result.
        """
        pass


class ActionProvider(abc.ABC):
    """Serves as a source of actions for the system.

    A concrete implementation of this interface can use its own
    way of delivering actions to the system. It can store actions
    in a database, get them over HTTP, AMQP or any other transport.
    Or it can simply provide a static collection of actions and keep
    them in memory throughout the cycle of the application.
    """

    def __init__(self, name):
        self._name = name

    @property
    def name(self):
        """The name of the action provider.

        Different action providers can use it differently.
        Some may completely ignore it, others may use it, for example,
        for searching actions in a certain way.
        """
        return self._name

    @abc.abstractmethod
    def find(self, action_name, namespace=None):
        """Finds action descriptor by name.

        :param action_name: Action name.
        :param namespace: Action namespace. None is used for the default
            namespace.
        :return: An instance of ActionDescriptor or None, if not found.
        """
        pass

    @abc.abstractmethod
    def find_all(self, namespace=None, limit=None, sort_fields=None,
                 sort_dirs=None, filters=None):
        """Finds all action descriptors for this provider.

        :param namespace: Optional. Action namespace.
        :param limit: Positive integer or None. Maximum number of action
            descriptors to return in a single result.
        :param sort_fields: Optional. A list of action descriptor fields
            that define sorting of the result set.
        :param sort_dirs: Optional. A list of sorting orders ("asc" or "desc")
            in addition to the "sort_fields" argument.
        :param filters: Optional. A dictionary describing AND-joined filters.
        :return: List of ActionDescriptor instances.
        """
        pass



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from mistral_lib import actions
from mistral_lib.actions.providers import python
from mistral_lib.tests import base as tests_base


class HelloAction(actions.Action):
    """I help with testing."""

    def __init__(self, f_name, l_name):
        super(HelloAction, self).__init__()

        self._f_name = f_name
        self._l_name = l_name

    def run(self, context):
        return 'Hello %s %s!' % (self._f_name, self._l_name)


class TestActionProvider(actions.ActionProvider):
    def __init__(self, name):
        super(TestActionProvider, self).__init__(name)

        self.action_descs = {}

    def add_action_descriptor(self, action_desc):
        self.action_descs[action_desc.name] = action_desc

    def find(self, action_name, namespace=None):
        return self.action_descs.get(action_name)

    def find_all(self, namespace=None, limit=None, sort_fields=None,
                 sort_dirs=None, **filters):
        return self.action_descs.values()


class TestActionProviders(tests_base.TestCase):
    def test_python_action_descriptor(self):
        action_desc = python.PythonActionDescriptor('test_action', HelloAction)

        # Check descriptor attributes.
        self.assertEqual('test_action', action_desc.name)
        self.assertEqual(HelloAction, action_desc.action_class)
        self.assertEqual(
            HelloAction.__module__ + '.' + HelloAction.__name__,
            action_desc.action_class_name
        )
        self.assertIsNone(action_desc.action_class_attributes)
        self.assertEqual('I help with testing.', action_desc.description)
        self.assertEqual('f_name, l_name', action_desc.params_spec)

        # Instantiate the action and check how it works.
        action = action_desc.instantiate(
            {'f_name': 'Jhon', 'l_name': 'Doe'},
            {}
        )

        res = action.run(None)

        self.assertEqual('Hello Jhon Doe!', res)

    def test_composite_action_provider(self):
        # Check empty provider.
        composite_provider = actions.CompositeActionProvider('test', [])

        self.assertEqual(0, len(composite_provider.find_all()))

        # Add two test providers.
        provider1 = TestActionProvider('provider1')

        action_desc1 = python.PythonActionDescriptor('action1', HelloAction)
        action_desc2 = python.PythonActionDescriptor('action2', HelloAction)
        action_desc3 = python.PythonActionDescriptor('action3', HelloAction)
        action_desc4 = python.PythonActionDescriptor('action4', HelloAction)

        provider1.add_action_descriptor(action_desc1)
        provider1.add_action_descriptor(action_desc2)

        provider2 = TestActionProvider('provider2')

        provider2.add_action_descriptor(action_desc3)
        provider2.add_action_descriptor(action_desc4)

        composite_provider = actions.CompositeActionProvider(
            'test',
            [provider1, provider2]
        )

        self.assertEqual(4, len(composite_provider.find_all()))

        self.assertEqual(action_desc1, composite_provider.find('action1'))
        self.assertEqual(action_desc2, composite_provider.find('action2'))
        self.assertEqual(action_desc3, composite_provider.find('action3'))
        self.assertEqual(action_desc4, composite_provider.find('action4'))



# Copyright 2017 - Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from mistral_lib.actions.base import Action
from mistral_lib.actions.base import ActionDescriptor
from mistral_lib.actions.base import ActionProvider
from mistral_lib.actions.providers.composite import CompositeActionProvider
from mistral_lib.actions.providers.python import PythonActionDescriptor
from mistral_lib.actions.types import Result

__all__ = [
    'Action',
    'Result',
    'ActionDescriptor',
    'ActionProvider',
    'PythonActionDescriptor',
    'CompositeActionProvider'
]



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import warnings

from mistral_lib import serialization


class ActionContext(serialization.MistralSerializable):

    def __init__(self, security_ctx=None, execution_ctx=None):
        self.security = security_ctx
        self.execution = execution_ctx

    def _deprecation_warning(self, name):
        warnings.warn(
            "context.{0} is deprecated from the context passed to actions. "
            "Please use context.security.{0}. It will be removed in a future "
            "release.", DeprecationWarning
        )

    def __getattribute__(self, name):
        deprecated = [
            "auth_cacert", "auth_token", "auth_uri", "expires_at", "insecure",
            "is_target", "is_trust_scoped", "project_id", "project_name",
            "redelivered", "region_name", "service_catalog", "trust_id",
            "user_name"
        ]

        if name in deprecated:
            self._deprecation_warning(name)

            return getattr(self.security, name)

        return super(ActionContext, self).__getattribute__(name)


class SecurityContext(object):
    def __init__(self, auth_uri=None, auth_cacert=None, insecure=None,
                 service_catalog=None, region_name=None, is_trust_scoped=None,
                 redelivered=None, expires_at=None, trust_id=None,
                 is_target=None, project_id=None, project_name=None,
                 user_name=None, auth_token=None):
        self.auth_uri = auth_uri
        self.auth_cacert = auth_cacert
        self.insecure = insecure
        self.service_catalog = service_catalog
        self.region_name = region_name
        self.is_trust_scoped = is_trust_scoped
        self.redelivered = redelivered
        self.expires_at = expires_at
        self.trust_id = trust_id
        self.is_target = is_target
        self.project_id = project_id
        self.project_name = project_name
        self.user_name = user_name
        self.auth_token = auth_token


class ExecutionContext(object):
    def __init__(self, workflow_execution_id=None, task_execution_id=None,
                 action_execution_id=None, workflow_name=None,
                 callback_url=None, task_id=None, with_items_index=0,
                 task_rerun_no=0, task_rerun_id=None,
                 workflow_propagated_headers=None):
        self.workflow_execution_id = workflow_execution_id
        self.task_execution_id = task_execution_id
        self.action_execution_id = action_execution_id
        self.workflow_name = workflow_name
        self.callback_url = callback_url
        self.with_items_index = with_items_index
        self.task_rerun_no = task_rerun_no
        self.task_rerun_id = task_rerun_id
        self.workflow_propagated_headers = workflow_propagated_headers

        if task_id is not None:
            self.task_execution_id = task_id
            self._deprecate_task_id_warning()

    def _deprecate_task_id_warning(self):
        warnings.warn(
            "context.execution.task_id was deprecated in the Queens cycle. "
            "Please use context.execution.task_execution_id. It will be "
            "removed in a future release.", DeprecationWarning
        )

    @property
    def task_id(self):
        self._deprecate_task_id_warning()

        return self.task_execution_id


class ActionContextSerializer(serialization.DictBasedSerializer):
    def serialize_to_dict(self, entity):
        return {
            'security': vars(entity.security),
            'execution': vars(entity.execution),
        }

    def deserialize_from_dict(self, entity_dict):
        return ActionContext(
            security_ctx=SecurityContext(**entity_dict['security']),
            execution_ctx=ExecutionContext(**entity_dict['execution'])
        )


serialization.register_serializer(ActionContext, ActionContextSerializer())



# Copyright 2020 Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral_lib.actions.providers import base
from mistral_lib.utils import inspect_utils as i_utils


class PythonActionDescriptor(base.ActionDescriptorBase):
    def __init__(self, name, action_cls, action_cls_attrs=None, namespace=None,
                 project_id=None, scope=None):
        super(PythonActionDescriptor, self).__init__(
            name,
            i_utils.get_docstring(action_cls),
            i_utils.get_arg_list_as_str(action_cls.__init__),
            namespace,
            project_id,
            scope
        )

        self._action_cls = action_cls
        self._action_cls_attrs = action_cls_attrs

    def __repr__(self):
        return 'Python action [name=%s, cls=%s]' % (
            self.name,
            self._action_cls
        )

    def instantiate(self, params, wf_ctx):
        if not self._action_cls_attrs:
            # No need to create new dynamic type.
            return self._action_cls(**params)

        dynamic_cls = type(
            self._action_cls.__name__,
            (self._action_cls,),
            self._action_cls_attrs
        )

        return dynamic_cls(**params)

    @property
    def action_class(self):
        return self._action_cls

    @property
    def action_class_name(self):
        return "{}.{}".format(
            self._action_cls.__module__,
            self._action_cls.__name__
        )

    @property
    def action_class_attributes(self):
        return self._action_cls_attrs



# Copyright 2013 - Mirantis, Inc.
# Copyright 2017 - Nokia Networks.
# Copyright 2017 - Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

from yaql.language import utils as yaql_utils

from mistral_lib.tests import base as tests_base
from mistral_lib import utils

import testtools.matchers as ttm


LEFT = {
    'key1': {
        'key11': "val11"
    },
    'key2': 'val2'
}

RIGHT = {
    'key1': {
        'key11': "val111111",
        'key12': "val12",
        'key13': {
            'key131': 'val131'
        }
    },
    'key2': 'val2222',
    'key3': 'val3'
}


class TestUtils(tests_base.TestCase):

    def test_merge_dicts(self):
        left = copy.deepcopy(LEFT)
        right = copy.deepcopy(RIGHT)

        expected = {
            'key1': {
                'key11': "val111111",
                'key12': "val12",
                'key13': {
                    'key131': 'val131'
                }
            },
            'key2': 'val2222',
            'key3': 'val3'
        }

        utils.merge_dicts(left, right)

        self.assertDictEqual(left, expected)

    def test_merge_dicts_overwrite_false(self):
        left = copy.deepcopy(LEFT)
        right = copy.deepcopy(RIGHT)

        expected = {
            'key1': {
                'key11': "val11",
                'key12': "val12",
                'key13': {
                    'key131': 'val131'
                }
            },
            'key2': 'val2',
            'key3': 'val3'
        }

        utils.merge_dicts(left, right, overwrite=False)

        self.assertDictEqual(left, expected)

    def test_itersubclasses(self):
        class A(object):
            pass

        class B(A):
            pass

        class C(A):
            pass

        class D(C):
            pass

        self.assertEqual([B, C, D], list(utils.iter_subclasses(A)))

    def test_get_dict_from_entries(self):
        input = ['param1', {'param2': 2}]
        input_dict = utils.get_dict_from_entries(input)

        self.assertIn('param1', input_dict)
        self.assertIn('param2', input_dict)
        self.assertEqual(2, input_dict.get('param2'))
        self.assertIs(input_dict.get('param1'), utils.NotDefined)

    def test_get_input_dict_from_string(self):
        self.assertDictEqual(
            {
                'param1': utils.NotDefined,
                'param2': 2,
                'param3': 'var3'
            },
            utils.get_dict_from_string('param1, param2=2, param3="var3"')
        )

        self.assertDictEqual({}, utils.get_dict_from_string(''))

    def test_cut_string(self):
        s = 'Hello, Mistral!'

        self.assertEqual('Hello...', utils.cut_string(s, length=5))
        self.assertEqual(s, utils.cut_string(s, length=100))
        self.assertEqual(s, utils.cut_string(s, length=-1))

    def test_cut_list(self):
        list_ = ['Hello, Mistral!', 'Hello, OpenStack!']

        self.assertEqual("['Hello, M...", utils.cut_list(list_, 13))
        self.assertEqual("['Hello, Mistr...", utils.cut_list(list_, 17))
        self.assertEqual(
            "['Hello, Mistral!', 'He...",
            utils.cut_list(list_, 26))

        self.assertEqual(
            "['Hello, Mistral!', 'Hello, OpenStack!']",
            utils.cut_list(list_, 100)
        )

        self.assertEqual(
            "['Hello, Mistral!', 'Hello, OpenStack!']",
            utils.cut_list(list_, -1)
        )

        self.assertEqual("[1, 2...", utils.cut_list([1, 2, 3, 4, 5], 8))
        self.assertEqual("[1, 2,...", utils.cut_list([1, 2, 3, 4, 5], 9))
        self.assertEqual("[1, 2, 3...", utils.cut_list([1, 2, 3, 4, 5], 11))

        self.assertRaises(ValueError, utils.cut_list, (1, 2))

    def test_cut_list_with_large_dict_of_str(self):
        d = [str(i) for i in range(65535)]
        s = utils.cut(d, 65535)
        self.assertThat(len(s), ttm.Not(ttm.GreaterThan(65535)))

    def test_cut_list_with_large_dict_of_int(self):
        d = [i for i in range(65535)]
        s = utils.cut(d, 65535)
        self.assertThat(len(s), ttm.Not(ttm.GreaterThan(65535)))

    def test_cut_list_with_large_dict_of_dict(self):
        d = [{'value': str(i)} for i in range(65535)]
        s = utils.cut(d, 65535)
        self.assertThat(len(s), ttm.Not(ttm.GreaterThan(65535)))

    def test_cut_list_for_state_info(self):
        d = [{'value': 'This is a string that exceeds 35 characters'}
             for i in range(2000)]
        s = utils.cut(d, 65500)
        self.assertThat(len(s), ttm.Not(ttm.GreaterThan(65500)))

    def test_cut_dict_with_strings(self):
        d = {'key1': 'value1', 'key2': 'value2'}

        s = utils.cut_dict(d, 13)

        self.assertIn(s, ["{'key1': '...", "{'key2': '..."])

        s = utils.cut_dict(d, 15)

        self.assertIn(s, ["{'key1': 'va...", "{'key2': 'va..."])

        s = utils.cut_dict(d, 22)

        self.assertIn(
            s,
            ["{'key1': 'value1', ...", "{'key2': 'value2', ..."]
        )

        self.assertIn(
            utils.cut_dict(d, 100),
            [
                "{'key1': 'value1', 'key2': 'value2'}",
                "{'key2': 'value2', 'key1': 'value1'}"
            ]
        )

        self.assertIn(
            utils.cut_dict(d, -1),
            [
                "{'key1': 'value1', 'key2': 'value2'}",
                "{'key2': 'value2', 'key1': 'value1'}"
            ]
        )

        self.assertRaises(ValueError, utils.cut_dict, (1, 2))

    def test_cut_dict_with_digits(self):
        d = {1: 2, 3: 4}

        s = utils.cut_dict(d, 10)

        self.assertIn(s, ["{1: 2, ...", "{3: 4, ..."])

        s = utils.cut_dict(d, 11)

        self.assertIn(s, ["{1: 2, 3...", "{3: 4, 1..."])

        s = utils.cut_dict(d, 100)

        self.assertIn(s, ["{1: 2, 3: 4}", "{3: 4, 1: 2}"])

    def test_cut_dict_with_large_dict_of_str(self):
        d = {}
        for i in range(65535):
            d[str(i)] = str(i)
        s = utils.cut(d, 65535)
        self.assertThat(len(s), ttm.Not(ttm.GreaterThan(65535)))

    def test_cut_dict_with_large_dict_of_int(self):
        d = {}
        for i in range(65535):
            d[i] = i
        s = utils.cut(d, 65535)
        self.assertThat(len(s), ttm.Not(ttm.GreaterThan(65535)))

    def test_cut_dict_with_large_dict_of_dict(self):
        d = {}
        for i in range(65535):
            d[i] = {'value': str(i)}
        s = utils.cut(d, 65535)
        self.assertThat(len(s), ttm.Not(ttm.GreaterThan(65535)))

    def test_cut_dict_for_state_info(self):
        d = {}
        for i in range(2000):
            d[i] = {'value': 'This is a string that exceeds 35 characters'}
        s = utils.cut(d, 65500)
        self.assertThat(len(s), ttm.Not(ttm.GreaterThan(65500)))

    def test_mask_data(self):
        payload = {'adminPass': 'fooBarBaz'}
        expected = {'adminPass': '***'}
        self.assertEqual(expected, utils.mask_data(payload))

        payload = """adminPass='fooBarBaz'"""
        expected = """adminPass='***'"""
        self.assertEqual(expected, utils.mask_data(payload))

        payload = [{'adminPass': 'fooBarBaz'}, {"new_pass": "blah"}]
        expected = [{'adminPass': '***'}, {"new_pass": "***"}]
        self.assertEqual(expected, utils.mask_data(payload))

        payload = ["adminPass", 'fooBarBaz']
        expected = ["adminPass", 'fooBarBaz']
        self.assertEqual(expected, utils.mask_data(payload))

    def test_json_serialize_frozen_dict(self):
        data = yaql_utils.FrozenDict(a=1, b=2, c=iter([1, 2, 3]))

        json_str = utils.to_json_str(data)

        self.assertIsNotNone(json_str)

        self.assertIn('"a": 1', json_str)
        self.assertIn('"b": 2', json_str)
        self.assertIn('"c": [1, 2, 3]', json_str)

    def test_json_serialize_generator(self):
        def _list_stream(_list):
            for i in _list:
                yield i

        gen = _list_stream(
            [1, yaql_utils.FrozenDict(a=1), _list_stream([12, 15])]
        )

        self.assertEqual('[1, {"a": 1}, [12, 15]]', utils.to_json_str(gen))

    def test_json_serialize_dict_of_generators(self):
        def _f(cnt):
            for i in range(1, cnt + 1):
                yield i

        data = {'numbers': _f(3)}

        self.assertEqual('{"numbers": [1, 2, 3]}', utils.to_json_str(data))

    def test_json_serialize_range(self):
        self.assertEqual("[1, 2, 3, 4]", utils.to_json_str(range(1, 5)))

    def test_json_serialize_iterator_of_frozen_dicts(self):
        data = iter(
            [
                yaql_utils.FrozenDict(a=1, b=2, c=iter([1, 2, 3])),
                yaql_utils.FrozenDict(
                    a=11,
                    b=yaql_utils.FrozenDict(b='222'),
                    c=iter(
                        [
                            1,
                            yaql_utils.FrozenDict(
                                a=iter([4, yaql_utils.FrozenDict(a=99)])
                            )
                        ]
                    )
                )
            ]
        )

        json_str = utils.to_json_str(data)

        self.assertIsNotNone(json_str)

        # Checking the first item.
        self.assertIn('"a": 1', json_str)
        self.assertIn('"b": 2', json_str)
        self.assertIn('"c": [1, 2, 3]', json_str)

        # Checking the first item.
        self.assertIn('"a": 11', json_str)
        self.assertIn('"b": {"b": "222"}', json_str)
        self.assertIn('"c": [1, {"a": [4, {"a": 99}]}]', json_str)



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from mistral_lib import actions
from mistral_lib.actions import context
from mistral_lib.tests import base as tests_base


class TestAction(actions.Action):

    def run(self, context):
        return context


class TestActionsBase(tests_base.TestCase):

    def test_run_empty_context(self):
        ctx = context.ActionContext()
        action = TestAction()
        result = action.run(ctx)
        assert result == ctx



# Copyright 2014 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import time

from mistral_lib import actions
from mistral_lib.tests import base
from mistral_lib.utils import inspect_utils as i_u


class DummyHTTPAction(actions.Action):
    def __init__(self,
                 url,
                 method="GET",
                 params=None,
                 body=None,
                 headers=None,
                 cookies=None,
                 auth=None,
                 timeout=None,
                 allow_redirects=None,
                 proxies=None,
                 verify=None):
        super(DummyHTTPAction, self).__init__()

    def run(self, context):
        return context


class DummyRunTask(object):
    def __init__(self, wf_ex, wf_spec, task_spec, ctx, triggered_by=None,
                 handles_error=False):
        pass


class ClassWithProperties(object):

    a = 1

    @property
    def prop(self):
        pass


class InspectUtilsTest(base.TestCase):
    def test_get_parameters_str(self):
        action_class = DummyHTTPAction
        parameters_str = i_u.get_arg_list_as_str(action_class.__init__)

        http_action_params = (
            'url, method="GET", params=null, body=null, '
            'headers=null, cookies=null, auth=null, '
            'timeout=null, allow_redirects=null, '
            'proxies=null, verify=null'
        )

        self.assertEqual(http_action_params, parameters_str)

    def test_get_parameters_str_all_mandatory(self):
        clazz = DummyRunTask
        parameters_str = i_u.get_arg_list_as_str(clazz.__init__)

        self.assertEqual(
            'wf_ex, wf_spec, task_spec, ctx, triggered_by=null,'
            ' handles_error=false',
            parameters_str
        )

    def test_get_parameters_str_with_function_parameter(self):

        def test_func(foo, bar=None, test_func=time.sleep):
            pass

        parameters_str = i_u.get_arg_list_as_str(test_func)

        self.assertEqual("foo, bar=null", parameters_str)

    def test_get_public_fields(self):

        attrs = i_u.get_public_fields(ClassWithProperties)

        self.assertEqual(attrs, {'a': 1})



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from mistral_lib.actions import context
from mistral_lib.tests import base as tests_base


def _fake_context():
    security_ctx = context.SecurityContext(
        auth_uri='auth_uri',
        auth_cacert='auth_cacert',
        insecure='insecure',
        service_catalog='service_catalog',
        region_name='region_name',
        is_trust_scoped='is_trust_scoped',
        redelivered='redelivered',
        expires_at='expires_at',
        trust_id='trust_id',
        is_target='is_target',
        project_id='project_id')

    execution_ctx = context.ExecutionContext(
        workflow_execution_id='workflow_execution_id',
        task_execution_id='task_execution_id',
        action_execution_id='action_execution_id',
        workflow_name='workflow_name',
        callback_url='callback_url',
        with_items_index=0,
        task_rerun_no=0,
        task_rerun_id='task_rerun_id')

    return context.ActionContext(security_ctx, execution_ctx)


class TestActionsBase(tests_base.TestCase):

    def test_empty_context(self):
        ctx = context.ActionContext(
            context.SecurityContext(),
            context.ExecutionContext()
        )

        self.assertIsInstance(ctx.security, context.SecurityContext)
        self.assertIsInstance(ctx.execution, context.ExecutionContext)

        self.assertEqual(ctx.security.auth_uri, None)
        self.assertEqual(ctx.execution.workflow_name, None)

    def test_deprecated_properties(self):
        ctx = _fake_context()

        deprecated_properties = [
            'auth_cacert', 'auth_token', 'auth_uri', 'expires_at',
            'insecure', 'is_target', 'is_trust_scoped', 'project_id',
            'project_name', 'redelivered', 'region_name', 'service_catalog',
            'trust_id', 'user_name'
        ]

        for deprecated in deprecated_properties:
            old = getattr(ctx, deprecated)
            new = getattr(ctx.security, deprecated)

            self.assertEqual(old, new)


class TestActionContextSerializer(tests_base.TestCase):

    def test_serialization(self):
        ctx = _fake_context()

        serialiser = context.ActionContextSerializer()

        dict_ctx = serialiser.serialize_to_dict(ctx)

        self.assertEqual(dict_ctx['security'], vars(ctx.security))
        self.assertEqual(dict_ctx['execution'], vars(ctx.execution))

    def test_deserialization(self):
        ctx = _fake_context()

        serialiser = context.ActionContextSerializer()

        dict_ctx = serialiser.serialize_to_dict(ctx)
        ctx_2 = serialiser.deserialize_from_dict(dict_ctx)

        self.assertEqual(ctx.security.auth_uri, ctx_2.security.auth_uri)
        self.assertEqual(
            ctx.execution.workflow_name,
            ctx_2.execution.workflow_name
        )



#    Copyright 2017 Nokia Networks.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from oslo_serialization import jsonutils


_SERIALIZER = None


class Serializer(object):
    """Base interface for entity serializers.

    A particular serializer knows how to convert a certain object
    into a string and back from that string into an object whose
    state is equivalent to the initial object.
    """

    @abc.abstractmethod
    def serialize(self, entity):
        """Converts the given object into a string.

        :param entity: An object to be serialized.
        :return String containing the state of the object in serialized form.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def deserialize(self, data_str):
        """Converts the given string into an object.

        :param data_str: String containing the state of the object in
            serialized form.
        :return: An object.
        """
        raise NotImplementedError


class DictBasedSerializer(Serializer):
    """Dictionary-based serializer.

    It slightly simplifies implementing custom serializers by introducing
    a contract based on dictionary. A serializer class extending this class
    just needs to implement conversion from object into dict and from dict
    to object. It doesn't need to convert into string and back as required
    bye the base serializer contract. Conversion into string is implemented
    once with regard to possible problems that may occur for collection and
    primitive types as circular dependencies, correct date format etc.
    """

    def serialize(self, entity):
        if entity is None:
            return None

        entity_dict = self.serialize_to_dict(entity)

        return jsonutils.dumps(
            jsonutils.to_primitive(entity_dict, convert_instances=True)
        )

    def deserialize(self, data_str):
        if data_str is None:
            return None

        entity_dict = jsonutils.loads(data_str)

        return self.deserialize_from_dict(entity_dict)

    @abc.abstractmethod
    def serialize_to_dict(self, entity):
        raise NotImplementedError

    @abc.abstractmethod
    def deserialize_from_dict(self, entity_dict):
        raise NotImplementedError


class MistralSerializable(object):
    """A mixin to generate a serialization key for a custom object."""

    @classmethod
    def get_serialization_key(cls):
        return "%s.%s" % (cls.__module__, cls.__name__)


class PolymorphicSerializer(Serializer):
    """Polymorphic serializer.

    The purpose of this class is to serve as a serialization router
    between serializers that can work with entities of particular type.
    All concrete serializers associated with concrete entity classes
    should be registered via method 'register', after that an instance
    of polymorphic serializer can be used as a universal serializer
    for an RPC system or something else.
    When converting an object into a string this serializer also writes
    a special key into the result string sequence so that it's possible
    to find a proper serializer when deserializing this object.
    If a primitive value is given as an entity this serializer doesn't
    do anything special and simply converts a value into a string using
    jsonutils. Similar when it converts a string into a primitive value.
    """

    def __init__(self):
        # {serialization key: serializer}
        self.serializers = {}

    @staticmethod
    def _get_serialization_key(entity_cls):
        if issubclass(entity_cls, MistralSerializable):
            return entity_cls.get_serialization_key()

        return None

    def register(self, entity_cls, serializer):
        key = self._get_serialization_key(entity_cls)

        if not key:
            return

        if key in self.serializers:
            raise RuntimeError(
                "A serializer for the entity class has already been"
                " registered: %s" % entity_cls
            )

        self.serializers[key] = serializer

    def unregister(self, entity_cls):
        key = self._get_serialization_key(entity_cls)

        if not key:
            return

        if key in self.serializers:
            del self.serializers[key]

    def cleanup(self):
        self.serializers.clear()

    def serialize(self, entity):
        if entity is None:
            return None

        key = self._get_serialization_key(type(entity))

        # Primitive or not registered type.
        if not key:
            return jsonutils.dumps(
                jsonutils.to_primitive(entity, convert_instances=True)
            )

        serializer = self.serializers.get(key)

        if not serializer:
            raise RuntimeError(
                "Failed to find a serializer for the key: %s" % key
            )

        result = {
            '__serial_key': key,
            '__serial_data': serializer.serialize(entity)
        }

        return jsonutils.dumps(result)

    def deserialize(self, data_str):
        if data_str is None:
            return None

        data = jsonutils.loads(data_str)

        if isinstance(data, dict) and '__serial_key' in data:
            serializer = self.serializers.get(data['__serial_key'])

            return serializer.deserialize(data['__serial_data'])

        return data


def get_polymorphic_serializer():
    global _SERIALIZER

    if _SERIALIZER is None:
        _SERIALIZER = PolymorphicSerializer()

    return _SERIALIZER


def register_serializer(entity_cls, serializer):
    get_polymorphic_serializer().register(entity_cls, serializer)


def unregister_serializer(entity_cls):
    get_polymorphic_serializer().unregister(entity_cls)


def cleanup():
    get_polymorphic_serializer().cleanup()



# Copyright 2020 Nokia Software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import copy

from mistral_lib import actions
from mistral_lib import exceptions as exc
from mistral_lib import utils


def _compare_parameters(expected_params, actual_params):
    """Compares the expected parameters with the actual parameters.

    :param expected_params: Expected dict of parameters.
    :param actual_params: Actual dict of parameters.
    :return: Tuple {missing parameter names, unexpected parameter names}
    """

    missing_params = []
    unexpected_params = copy.deepcopy(list((actual_params or {}).keys()))

    for p_name, p_value in expected_params.items():
        if p_value is utils.NotDefined and p_name not in unexpected_params:
            missing_params.append(str(p_name))

        if p_name in unexpected_params:
            unexpected_params.remove(p_name)

    return missing_params, unexpected_params


class ActionDescriptorBase(actions.ActionDescriptor, abc.ABC):
    def __init__(self, name, desc, params_spec, namespace=None,
                 project_id=None, scope=None):
        self._name = name
        self._desc = desc
        self._params_spec = params_spec
        self._namespace = namespace
        self._project_id = project_id
        self._scope = scope

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._desc

    @property
    def params_spec(self):
        return self._params_spec

    @property
    def namespace(self):
        return self._namespace

    @property
    def project_id(self):
        return self._project_id

    @property
    def scope(self):
        return self._scope

    @property
    def action_class_name(self):
        return None

    @property
    def action_class_attributes(self):
        return None

    def check_parameters(self, params):
        # Don't validate action input if action initialization
        # method contains ** argument.
        if '**' in self.params_spec:
            return

        expected_params = utils.get_dict_from_string(self.params_spec)

        actual_params = params or {}

        missing, unexpected = _compare_parameters(
            expected_params,
            actual_params
        )

        if missing or unexpected:
            msg = 'Invalid input [name=%s, class=%s'
            msg_props = [self.name, self.action_class_name]

            if missing:
                msg += ', missing=%s'
                msg_props.append(missing)

            if unexpected:
                msg += ', unexpected=%s'
                msg_props.append(unexpected)

            msg += ']'

            raise exc.ActionException(msg % tuple(msg_props))

    def post_process_result(self, result):
        return result
